# -*- coding: utf-8 -*-
"""
Likelihoods that are mixtures of distributions.
"""
from functools import partial
from typing import Callable, Optional

import numpy as np
from scipy import stats

from .base import BaseModel, UniformPriorMixin


class MixtureOfDistributions(UniformPriorMixin, BaseModel):
    """Mixture of distributions.

    Available distributions: Gaussian, Uniform, Gamma, HalfNorm

    Parameters
    ----------
    distributions
        Dictionary with the names of the distributions as keys and the number
        of each distribution to add as the values.
    bounds
        Dictionary of priors bounds
    distributions_kwargs
        Dictionary of dictionaries where the key is the name of distribution
        and the values are a dictionary of keyword arguments.
    map_fn
        Map function use when computing the likelihood.
    """

    def __init__(
        self,
        distributions: Optional[dict] = None,
        bounds: Optional[dict] = None,
        distributions_kwargs: Optional[dict] = None,
        map_fn: Optional[Callable] = None,
    ) -> None:
        self.bounds_mapping = dict(
            gaussian=[-10.0, 10.0],
            uniform=[-5.0, 5.0],
            gamma=[0.0, 10.0],
            halfnorm=[0.0, 10.0],
        )

        if distributions is None:
            distributions = {"gaussian": 2, "uniform": 2, "gamma": 2}

        if distributions_kwargs:
            if set(distributions_kwargs).difference(self.bounds_mapping):
                raise ValueError(
                    "distribution_kwargs contains unknown distributions"
                )
        else:
            distributions_kwargs = {}

        self.base_mapping = dict(
            gaussian=stats.norm(
                **distributions_kwargs.get("gaussian", {})
            ).logpdf,
            uniform=stats.uniform(
                self.bounds_mapping["uniform"][0],
                np.ptp(self.bounds_mapping["uniform"]),
                **distributions_kwargs.get("uniform", {}),
            ).logpdf,
            gamma=stats.gamma(
                **distributions_kwargs.get("gamma", {"a": 1.99})
            ).logpdf,
            halfnorm=stats.halfnorm(
                **distributions_kwargs.get("halfnorm", {})
            ).logpdf,
        )

        self.mapping = dict()
        if bounds is None:
            bounds = dict()

        for dist, n in distributions.items():
            for i in range(n):
                name = dist + f"_{i}"
                self.mapping[name] = self.base_mapping[dist.lower()]
                if name not in bounds:
                    bounds[name] = self.bounds_mapping[dist]

        self.names = list(self.mapping.keys())
        self.bounds = bounds

        if map_fn is None:
            self.map_fn = map
        else:
            self.map_fn = map_fn

    @staticmethod
    def _log_likelihood_name(
        mapping: dict, x: np.ndarray, name: str
    ) -> np.ndarray:
        return mapping[name](x[name])

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Compute the log-likelihood.

        Parameters
        ----------
        x :
            Point or array of points as a structured array with fields
            that match the names of the model.

        Returns
        -------
        numpy.ndarray
            One-dimensional array of log-likelihoods
        """
        return np.vstack(
            list(
                self.map_fn(
                    partial(self._log_likelihood_name, self.mapping, x),
                    self.names,
                )
            )
        ).sum(axis=0)



"""Tests for signal-based models"""
import numpy as np
import pytest

from nessai_models.signals import (
    LinearSignal,
    SinusoidalSignal,
)

all_models = [
    LinearSignal,
    SinusoidalSignal,
]


@pytest.fixture(params=all_models)
def SignalModelClass(request):
    """Signal model classes fixture."""
    return request.param


@pytest.mark.parametrize("n_points", [10, 100])
@pytest.mark.parametrize("sigma", [1, 2])
def test_max_likelihood(SignalModelClass, n_points, sigma):
    """Assert the maximum likelihood is the correct value.

    Sets the data to have zero noise.
    """
    model = SignalModelClass(n_points=n_points, sigma=sigma)
    model.data = model.signal_model(**model.truth)
    expected = n_points * -np.log(2 * np.pi * sigma**2)
    actual = model.log_likelihood(model.truth)
    np.testing.assert_almost_equal(actual, expected, decimal=12)



# -*- coding: utf-8 -*-
"""
Gaussian mixture models.
"""
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
from scipy.stats import multivariate_normal, norm
from scipy.special import logsumexp

from .base import BaseModel, NDimensionalModel, UniformPriorMixin


class GaussianMixture(UniformPriorMixin, NDimensionalModel):
    """A Gaussian mixture model with n Gaussians defined on [-10, 10]^n.

    Parameters
    ----------
    dims : int
        Number of dimensions
    n_gaussians : int
        Number of Gaussians.
    weights : Optional[Union[Sequence[float], np.ndarray]]
        Weights for each of the Gaussian. Must sum to one.
    config : Optional[Union[List[Dict[str, numpy.ndarray]]]]
        List of configurations for each Gaussian. Each dictionary should have
        a mean and cov key.
    random_state : Optional[numpy.random.RandomState]
        Random state to use for generation configuration if `config` is not
        specified. If not specified `seed` is used instead.
    seed : int
        Random seed for seeding random number generation.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    """

    def __init__(
        self,
        dims: int = 4,
        n_gaussians: int = 2,
        weights: Optional[Union[Sequence[float], np.ndarray]] = None,
        config: Optional[List[Dict[str, np.ndarray]]] = None,
        random_state: Optional[np.random.RandomState] = None,
        seed: int = 1234,
        bounds: Union[Sequence[float], np.ndarray] = [-10.0, 10.0],
    ) -> None:
        super().__init__(dims, bounds)

        if random_state is None:
            random_state = np.random.RandomState(seed=seed)

        self.n_gaussians = n_gaussians
        if weights is None:
            self.weights = np.ones(n_gaussians) / n_gaussians
        else:
            if len(weights) != n_gaussians:
                ValueError("Length of weights must match number of Gaussians")
            self.weights = np.array(weights)
        self.gaussians = n_gaussians * [None]
        if config is None:
            config = n_gaussians * [None]
        elif len(config) != n_gaussians:
            raise RuntimeError("Config does not match number of Gaussians")

        for n in range(n_gaussians):
            if config[n] is None:
                config[n] = dict(
                    mean=random_state.uniform(bounds[0], bounds[1], dims),
                    cov=3 * random_state.rand() * np.eye(dims),
                )
            self.gaussians[n] = multivariate_normal(**config[n])

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Log-likelihood for the mixture of Gaussians."""
        _x = self.unstructured_view(x)
        b = np.broadcast_to(self.weights, (x.size, self.n_gaussians)).T
        arg = np.array([g.logpdf(_x) for g in self.gaussians])
        if arg.ndim == 1:
            arg = arg[:, np.newaxis]
        log_l = logsumexp(arg, b=b, axis=0)
        return log_l


class GaussianMixtureWithData(UniformPriorMixin, BaseModel):
    """
    A Gaussian mixture model with two peaks that uses samples and fits the
    the means standard deviations and weight.

    Based on the example from cpnest: \
        https://github.com/johnveitch/cpnest/blob/master/examples/gaussianmixture.py

    The parameters to estimate are the means, standard deviations and the
    weight.

    Parameters
    ----------
    n : int
        Number of data points to use.
    """

    def __init__(self, n: int = 1000) -> None:
        self.names = ["mu1", "sigma1", "mu2", "sigma2", "weight"]
        self.bounds = {
            "mu1": [-3, 3],
            "sigma1": [0.01, 1],
            "mu2": [-3, 3],
            "sigma2": [0.01, 1.0],
            "weight": [0.0, 1.0],
        }

        self.truth = {
            "mu1": 0.5,
            "sigma1": 0.5,
            "mu2": -1.5,
            "sigma2": 0.03,
            "weight": 0.2,
        }
        self.gaussian1 = norm(self.truth["mu1"], scale=self.truth["sigma1"])
        self.gaussian2 = norm(self.truth["mu2"], scale=self.truth["sigma2"])

        n1 = int(self.truth["weight"] * n)
        n2 = n - n1

        self.data = np.concatenate(
            [self.gaussian1.rvs(size=n1), self.gaussian2.rvs(size=n2)]
        )

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Returns log likelihood of given live point."""
        w = x["weight"][..., np.newaxis]
        mu1 = x["mu1"][..., np.newaxis]
        mu2 = x["mu2"][..., np.newaxis]
        sigma1 = x["sigma1"][..., np.newaxis]
        sigma2 = x["sigma2"][..., np.newaxis]
        log_l1 = np.sum(
            np.log(w)
            - np.log(sigma1)
            - 0.5 * ((self.data - mu1) / sigma1) ** 2,
            axis=-1,
        )
        log_l2 = np.sum(
            np.log(1.0 - w)
            - np.log(sigma2)
            - 0.5 * ((self.data - mu2) / sigma2) ** 2,
            axis=-1,
        )
        log_l = np.logaddexp(log_l1, log_l2)
        return log_l



# -*- coding: utf-8 -*-
"""Tests the base models from `nessai_models.base`."""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec

from nessai_models.base import (
    NDimensionalModel,
    UniformPriorMixin,
)


@pytest.mark.parametrize("bounds", [[-10.0, 10.0], np.array([-10.0, 10.0])])
def test_n_dimensional_model_bounds(bounds):
    """Test the n-dimensional model init."""
    model = create_autospec(NDimensionalModel)
    NDimensionalModel.__init__(model, 2, bounds)
    assert model.names == ["x_0", "x_1"]
    np.testing.assert_equal(model.bounds["x_0"], [-10.0, 10.0])
    np.testing.assert_equal(model.bounds["x_1"], [-10.0, 10.0])


def test_n_dimensional_model_bounds_invalid_type():
    """Assert an error is raised in the bounds are the incorrect type."""
    model = create_autospec(NDimensionalModel)
    with pytest.raises(TypeError) as excinfo:
        NDimensionalModel.__init__(model, 2, 10)
    assert "Invalid type" in str(excinfo.value)


def test_n_dimensional_model_bounds_invalid_length():
    """Assert an error is raised in the bounds are the incorrect length"""
    model = create_autospec(NDimensionalModel)
    with pytest.raises(ValueError) as excinfo:
        NDimensionalModel.__init__(model, 2, [1, 2, 3])
    assert "must have length 2" in str(excinfo.value)


def test_uniform_prior_mixin():
    """Assert the value returned by log-prior method is correct."""
    # Values won't be checked because of mocked method in_bounds
    x = np.zeros(10)
    lower_bounds = np.array([-10, -2, 2])
    upper_bounds = np.array([10, 1, 7])
    target = -np.log(20) - np.log(3) - np.log(5)
    obj = create_autospec(UniformPriorMixin)
    obj.in_bounds = MagicMock(return_value=np.ones(x.size).astype(bool))
    obj.lower_bounds = lower_bounds
    obj.upper_bounds = upper_bounds

    log_prob = UniformPriorMixin.log_prior(obj, x)

    obj.in_bounds.assert_called_once_with(x)
    np.testing.assert_equal(log_prob, target)


def test_uniform_prior_mixin_out_of_bounds():
    """Test the log-prior method when a point is deemed out of bounds"""
    # Values won't be checked because of mocked method in_bounds
    x = np.zeros(2)
    lower_bounds = np.array([-10, -2, 2])
    upper_bounds = np.array([10, 1, 7])
    target = np.array([-np.log(20) - np.log(3) - np.log(5), -np.inf])
    obj = create_autospec(UniformPriorMixin)
    obj.in_bounds = MagicMock(return_value=np.array([True, False]))
    obj.lower_bounds = lower_bounds
    obj.upper_bounds = upper_bounds

    log_prob = UniformPriorMixin.log_prior(obj, x)

    obj.in_bounds.assert_called_once_with(x)
    np.testing.assert_equal(log_prob, target)


def test_uniform_prior_mixin_to_unit_hypercube():
    """Assert samples are transformed to the correct range"""
    obj = create_autospec(UniformPriorMixin)
    obj.bounds = {"x": np.array([-10, 10])}
    obj.names = ["x"]

    x = np.array([(-10,), (0,), (10,)], dtype=[("x", "f8")])

    out = UniformPriorMixin.to_unit_hypercube(obj, x)

    np.testing.assert_array_equal(out["x"], np.array([0.0, 0.5, 1.0]))


def test_uniform_prior_mixin_from_unit_hypercube():
    """Assert samples are transformed to the correct range"""
    obj = create_autospec(UniformPriorMixin)
    obj.bounds = {"x": np.array([-10, 10])}
    obj.names = ["x"]

    x = np.array([(0.0,), (0.5,), (1,)], dtype=[("x", "f8")])

    out = UniformPriorMixin.from_unit_hypercube(obj, x)

    np.testing.assert_array_equal(out["x"], np.array([-10.0, 0.0, 10.0]))



# -*- coding: utf-8 -*-
"""
N-dimensional Gaussian likelihood
"""
from typing import Optional, Sequence, Union
import warnings

import numpy as np
from scipy.stats import multivariate_normal

from .base import NDimensionalModel, UniformPriorMixin


def compute_gaussian_ln_evidence(
    bounds: Union[list, tuple, np.ndarray],
    dims: Optional[int] = None,
) -> float:
    """Compute the ln-evidence for a unit Gaussian likelihood:

    Parameters
    ----------
    bounds : Union[list, tuple numpy.ndarray]
        Prior bounds. Either 1-d or 2-d. When 1-d input is given, `dims` must
        be specified and the prior bounds are assumed to be the same in each
        dimension.
    dims : Optional[int]
        Number of dimensions. If a 2-d bounds array is given, then the value
        of `dims` is checked against the shape of `bounds`.
    """
    if np.ndim(bounds) == 1:
        if dims is None:
            raise ValueError(
                "dims must be specified if bounds is 1-dimensional"
            )
        ln_z = -dims * np.log(np.ptp(bounds))
    else:
        if dims and np.shape(bounds)[0] != dims:
            raise ValueError(
                "When providing 2-d bounds and dims, dims must match the "
                "first dimension of bounds."
            )
        ln_z = -np.log(np.ptp(bounds, axis=-1)).sum()
    return ln_z


class Gaussian(UniformPriorMixin, NDimensionalModel):
    """A simple n-dimensional Guassian with uniform priors.

    Defaults to a unit Gaussian two dimensions and priors defined on
    [-10, 10]^n.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    mean : Optional[Sequence[float], numpy.ndarray]
        Mean of the Gaussian.
    cov: Optional[Sequence[float], numpy.ndarray]
        Covariance matrix of the Gaussian.
    normalise : bool
        If true, the log-likelihood will be renormalised such that the log-
        evidence is zero. Only applies when :code:`mean` and :code:`cov` are
        not specified.
    """

    def __init__(
        self,
        dims: int = 2,
        bounds: Union[Sequence[float], np.ndarray] = [-10.0, 10.0],
        mean: Union[Sequence[float], np.ndarray] = None,
        cov: Union[Sequence[float], np.ndarray] = None,
        normalise: bool = False,
    ) -> None:
        super().__init__(dims, bounds)

        self._norm_const = 0.0
        if mean is None:
            self.mean = np.zeros(self.dims)
        elif isinstance(mean, (float, int)):
            self.mean = float(mean) * np.ones(self.dims)
        else:
            self.mean = mean

        if cov is None:
            self.cov = np.eye(self.dims)
        else:
            self.cov = cov

        self.dist = multivariate_normal(mean=self.mean, cov=self.cov)
        self.normalise = normalise

        if cov is None and mean is None:
            self.ln_evidence = compute_gaussian_ln_evidence(
                bounds, dims=self.dims
            )
            if self.normalise:
                self._norm_const = self.ln_evidence
        else:
            self.ln_evidence = None
            if self.normalise:
                warnings.warn("Cannot normalise non-unit Gaussian")
                self.normalise = False

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Gaussian log-likelihood."""
        # Use a view rather than making a new copy of y
        return self.dist.logpdf(self.unstructured_view(x)) - self._norm_const



"""General configuration for all tests"""
from nessai_models import (
    Brewer,
    EggBox,
    Gaussian,
    GaussianMixture,
    GaussianMixtureWithData,
    HalfGaussian,
    LinearSignal,
    MixtureOfDistributions,
    Pyramid,
    Rosenbrock,
    SinusoidalSignal,
)
import pytest


all_models = [
    Brewer,
    EggBox,
    Gaussian,
    GaussianMixture,
    GaussianMixtureWithData,
    HalfGaussian,
    LinearSignal,
    MixtureOfDistributions,
    Pyramid,
    Rosenbrock,
    SinusoidalSignal,
]


@pytest.fixture(params=all_models)
def ModelClass(request):
    """Model classes fixture.

    Contains all implemented models.
    """
    return request.param



# -*- coding: utf-8 -*-
"""
Tests specific to the n-dimensional Gaussian Mixture models.
"""
from nessai.livepoint import numpy_array_to_live_points
import numpy as np
from scipy.stats import multivariate_normal
import pytest

from nessai_models.gaussianmixture import GaussianMixture


@pytest.mark.integration_test
def test_weighted_gm():
    """Integration test to verify the log-likelihood."""
    dims = 2
    n_points = 10
    weights = [0.1, 0.9]
    n_gaussians = 2
    config = [
        dict(mean=dims * [-1], cov=2 * np.eye(dims)),
        dict(mean=dims * [1], cov=2 * np.eye(dims)),
    ]
    model = GaussianMixture(
        dims=dims,
        weights=weights,
        config=config,
        n_gaussians=n_gaussians,
    )

    dist0 = multivariate_normal(
        mean=config[0]["mean"],
        cov=config[0]["cov"],
    )
    dist1 = multivariate_normal(
        mean=config[1]["mean"],
        cov=config[1]["cov"],
    )
    x = np.random.randn(n_points, dims)
    x_live = numpy_array_to_live_points(x, model.names)
    expected = np.log(weights[0] * dist0.pdf(x) + weights[1] * dist1.pdf(x))

    out = model.log_likelihood(x_live)
    assert out.shape == (n_points,)

    np.testing.assert_array_almost_equal_nulp(out, expected)



# -*- coding: utf-8 -*-
"""Tests for the Rosenbrock likelihood"""

from unittest.mock import MagicMock, create_autospec, patch

import numpy as np
import pytest

from nessai_models.rosenbrock import (
    Rosenbrock,
    rosenbrock,
    uncoupled_rosenbrock,
)


@pytest.fixture()
def model():
    return create_autospec(Rosenbrock)


@pytest.fixture(params=[rosenbrock, uncoupled_rosenbrock])
def func(request):
    return request.param


@pytest.mark.parametrize("n", [1, 2])
def test_rosenbrock_function(func, n):
    """Assert the output has the correct shape"""
    x = np.random.randn(n, 2)
    out = func(x)
    assert out.ndim == 1
    assert len(out) == n


def test_rosenbrock_function_1d(func):
    """Assert the output has the correct shape for a 1d input"""
    x = np.random.randn(2)
    out = func(x)
    assert out.ndim == 0


def test_rosenbrock_minimum(func):
    """Assert there is minimum at (1, 1).

    This should apply to both versions of the function.
    """
    np.testing.assert_equal(func(np.array([1, 1])), 0.0)


@pytest.mark.parametrize("uncoupled", [False, True])
def test_init(model, uncoupled):
    """Assert the model is instantiated correctly"""
    dims = 4
    bounds = [-1, 1]

    with patch("nessai_models.rosenbrock.NDimensionalModel.__init__") as mock:
        Rosenbrock.__init__(
            model, dims=dims, bounds=bounds, uncoupled=uncoupled
        )

    mock.assert_called_once_with(dims, bounds)

    if uncoupled:
        assert model._fn is uncoupled_rosenbrock
    else:
        assert model._fn is rosenbrock


def test_log_likelihood(model):
    """Assert the correct functions are called."""
    logL = 1.0
    model._fn = MagicMock(return_value=logL)
    model.unstructured_view = MagicMock(return_value="view")
    x = "input"
    out = Rosenbrock.log_likelihood(model, x)

    assert out == -logL
    model.unstructured_view.assert_called_once_with(x)
    model._fn.assert_called_once_with("view")



# -*- coding: utf-8 -*-
"""Test integration with nessai."""
import os
from nessai.flowsampler import FlowSampler
import pytest


@pytest.mark.slow_integration_test
def test_sampling(tmp_path, ModelClass):
    """Run a basic sampling run.

    Only runs 500 iterations since this should be enough to test the core
    functionality.
    """
    output = tmp_path / "test_sampling"
    output.mkdir()
    fs = FlowSampler(
        ModelClass(),
        nlive=100,
        poolsize=100,
        output=output,
        resume=False,
        max_iteration=500,
        stopping=1.0,
        plot=False,
    )
    fs.run(plot=False)
    # Make sure a result is produced
    os.path.exists(os.path.join(output, "result.json"))



"""Tests the for the mixture module."""
import multiprocessing.dummy as mp
from nessai_models.mixture import MixtureOfDistributions
import numpy as np

import pytest


@pytest.mark.integration_test
def test_multiprocessing_pool():
    """Assert the map function fom Pool can be used."""
    pool = mp.Pool(2)
    model = MixtureOfDistributions(map_fn=pool.map)
    x = model.new_point(100)
    model.log_likelihood(x)
    pool.close()


@pytest.mark.parametrize(
    "distributions",
    [{"gaussian": 2}, {"gamma": 2}, {"uniform": 2}, {"halfnorm": 2}],
)
def test_distributions(distributions):
    """Test the init and likelihood with different distributions"""
    model = MixtureOfDistributions(distributions=distributions)
    x = model.new_point()
    out = model.log_likelihood(x)
    assert np.isfinite(out).all()



# -*- coding: utf-8 -*-
"""
Tests specific to the n-dimensional Gaussian.
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec, patch

from nessai_models import Gaussian
from nessai_models.gaussian import compute_gaussian_ln_evidence


@pytest.fixture
def model():
    return create_autospec(Gaussian)


def test_init(model):
    """Test the init method"""
    model.dims = 2
    with patch("nessai_models.base.NDimensionalModel.__init__") as m, patch(
        "nessai_models.gaussian.compute_gaussian_ln_evidence", return_value=1.0
    ) as m1:
        Gaussian.__init__(model, 2, [-5, 5], normalise=True)
    m.assert_called_once_with(2, [-5, 5])
    m1.assert_called_once_with([-5, 5], dims=2)
    assert model.ln_evidence == 1.0
    assert model._norm_const == 1.0


def test_init_normalise_false(model):
    """Assert the normalisation constant is zero when normalise is False"""
    model.dims = 2
    Gaussian.__init__(model, 2, [-5, 5], normalise=False, cov=None, mean=None)
    assert model._norm_const == 0.0


def test_init_cannot_normalise(model):
    """Assert the normalisation constant is zero if normalisation is not
    possible.
    """
    mean = 1
    cov = np.eye(2)
    model.dims = 2
    with pytest.warns(Warning) as warninfo:
        Gaussian.__init__(
            model, 2, [-5, 5], mean=mean, cov=cov, normalise=True
        )
    assert model._norm_const == 0.0
    assert "Cannot normalise" in str(warninfo[0].message)


@pytest.mark.parametrize(
    "mean, expected",
    [
        (None, np.zeros(2)),
        (2.0, np.array([2, 2])),
        (np.array([2, 2]), np.array([2, 2])),
    ],
)
def test_init_mean(model, mean, expected):
    """Assert the mean is set correctly"""
    model.dims = len(expected)
    Gaussian.__init__(model, len(expected), [-5, 5], mean=mean)
    np.testing.assert_equal(model.mean, expected)


@pytest.mark.parametrize(
    "cov, expected",
    [
        (None, np.eye(2)),
        (np.eye(2), np.eye(2)),
    ],
)
def test_init_cov(model, cov, expected):
    """Assert the covariance is set correctly"""
    model.dims = len(expected)
    Gaussian.__init__(model, len(expected), [-5, 5], cov=cov)
    np.testing.assert_equal(model.cov, expected)


def test_log_likelihood(model):
    """Test the log-likelihood"""
    x = np.random.randn(4, 2)
    x_view = np.random.randn(4, 2)
    log_l = np.array([1, 2, 3, 4])
    model.unstructured_view = MagicMock(return_value=x_view)
    model.dist = MagicMock()
    model.dist.logpdf = MagicMock(return_value=log_l)
    model._norm_const = 1

    out = Gaussian.log_likelihood(model, x)
    np.testing.assert_equal(out, np.array([0, 1, 2, 3]))

    model.unstructured_view.assert_called_once_with(x)
    model.dist.logpdf.assert_called_once_with(x_view)


@pytest.mark.parametrize(
    "dims, bounds, expected",
    [
        [2, [-10, 10], -2 * np.log(20)],
        [None, [[-10, 10], [-10, 10]], -2 * np.log(20)],
        [4, [-5, 5], -4 * np.log(10)],
    ],
)
def test_gaussian_ln_evidence(dims, bounds, expected):
    """Assert the correct log evidence is returned."""
    out = compute_gaussian_ln_evidence(bounds, dims)
    assert expected == out


def test_gaussian_ln_evidence_dims_error():
    """Assert an error is raised if the bounds are 1-d and dims is None."""
    with pytest.raises(ValueError) as excinfo:
        compute_gaussian_ln_evidence([-5, 5])
    assert "dims must be specified" in str(excinfo.value)


def test_gaussian_ln_evidence_dims_wrong():
    """Assert an error is raised if the dims and bounds do not agree."""
    with pytest.raises(ValueError) as excinfo:
        compute_gaussian_ln_evidence(np.array([[-5, 5], [-5, 5]]), 3)
    assert "dims must match the first dimension" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Models for use with the nested sampler \
    `nessai <https://github.com/mj-will/nessai-models>`_.
"""
try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # for Python < 3.8
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

from .brewer import Brewer
from .eggbox import EggBox
from .gaussian import Gaussian
from .gaussianmixture import GaussianMixture, GaussianMixtureWithData
from .halfgaussian import HalfGaussian
from .mixture import MixtureOfDistributions
from .pyramid import Pyramid
from .rosenbrock import Rosenbrock
from .signals import LinearSignal, SinusoidalSignal

__all__ = [
    "Brewer",
    "EggBox",
    "Gaussian",
    "GaussianMixture",
    "GaussianMixtureWithData",
    "HalfGaussian",
    "LinearSignal",
    "MixtureOfDistributions",
    "Pyramid",
    "Rosenbrock",
    "SinusoidalSignal",
]



# -*- coding: utf-8 -*-
"""
N-dimensional egg box likelihood as defined in Feroz et al. 2008.
"""
from typing import Sequence, Union

import numpy as np
from .base import NDimensionalModel, UniformPriorMixin


class EggBox(UniformPriorMixin, NDimensionalModel):
    """N-dimensional egg box based of the version in \
        `Feroz et al. 2008 <https://arxiv.org/abs/0809.3437>`_

    This model extends the two-dimensional version of the Egg Box defined in
    Feroz at al. 2008 to n-dimensions as:

    .. math::

        \\log \\mathcal{L} =
        \\left[ 2 + \\prod_{i=1}^{n} \\cos \\frac{ \\theta_i}{2} \\right] ^ {5}

    Parameters
    ----------
    dims :
        Number of dimensions.
    bounds :
        Prior bounds.
    """

    def __init__(
        self,
        dims: int = 2,
        bounds: Union[Sequence[int], np.ndarray] = [0, 10.0 * np.pi],
    ) -> None:
        super().__init__(dims, bounds)

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Compute the log-likelihood.

        Parameters
        ----------
        x :
            Point or array of points as a structured array with fields
            that match the names of the model.

        Returns
        -------
        numpy.ndarray
            One-dimensional array of log-likelihoods
        """
        log_l = np.ones(x.size)
        for n in self.names:
            log_l += np.cos(x[n] / 2.0)
        return (log_l + 2.0) ** 5.0



# -*- coding: utf-8 -*-
"""
Likelihood described in Brewer et al. arXiv:0912.2380
"""
from typing import Sequence, Union

import numpy as np
from scipy.stats import multivariate_normal

from .base import NDimensionalModel, UniformPriorMixin


class Brewer(UniformPriorMixin, NDimensionalModel):
    """Bimodal likelihood described in Brewer et al. arXiv:0912.2380.

    Defaults match those described in the paper, but can be changed to match
    Skilling's original "Statistics Problem".

    Parameters
    ----------
    v_mean : float
        Mean of the wider peak.
    u_mean : float
        Mean of the narrower peak.
    v_width : float
        Width of the wider peak.
    u_width : float
        Width of the narrow peak.
    weight : float
        Relative weight of the narrower peak compared to the wider peak.
    dims : int
        Number of dimensions.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    """

    def __init__(
        self,
        v_mean: float = 0.0,
        u_mean: float = 0.031,
        v_width: float = 0.1,
        u_width: float = 0.01,
        weight: float = 100.0,
        dims: int = 20,
        bounds: Union[Sequence[float], np.ndarray] = [-0.5, 0.5],
    ) -> None:
        super().__init__(dims=dims, bounds=bounds)

        self.weight = weight
        self.ln_weight = np.log(weight)
        self.v_dist = multivariate_normal(
            mean=v_mean * np.ones(self.dims),
            cov=(v_width**2) * np.eye(self.dims),
        )
        self.u_dist = multivariate_normal(
            mean=u_mean * np.ones(self.dims),
            cov=(u_width**2) * np.eye(self.dims),
        )

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Likelihood function.

        Parameters
        ----------
        x : numpy.ndarray
            Structured array of parameters.

        Returns
        -------
        numpy.ndarray
            Array of log-likelihood values.
        """
        x = self.unstructured_view(x)
        return np.logaddexp(
            self.v_dist.logpdf(x),
            self.ln_weight + self.u_dist.logpdf(x),
        )



# -*- coding: utf-8 -*-
"""Example of using nessai_models with nessai.

This example uses a two-dimensional Gaussian Mixture model.
"""
from nessai.flowsampler import FlowSampler
from nessai.utils import setup_logger
from nessai_models import GaussianMixture

# Set up the logger as normal for nessai
setup_logger()

# The GaussianMixture model from nessai_models already has a parameter names,
# prior bounds, the log-prior and log-likelihood defined. So all we need to
# do is create an instance of class and, optionally, specify the number of
# dimensions.
# This particular model as has extra parameters that default values, for
# example the number of Gaussian in the mixture. By default, there weights and
# parameters (mean and covariance) will be randomised.
model = GaussianMixture(dims=2)

# We then create an instance of the sampler as usual
fs = FlowSampler(model, output="outdir", resume=False)
# And run the sampler
fs.run()



# -*- coding: utf-8 -*-
"""
N-dimensional Rosenbrock likelihood
"""
from typing import Sequence, Union

import numpy as np

from .base import NDimensionalModel, UniformPriorMixin


def uncoupled_rosenbrock(x: np.ndarray) -> np.ndarray:
    r"""Uncoupled Rosenbrock function in N dimensions.

    This is the simpler version which is the sum of N/2 uncouple 2D
    Rosenbrocks given by

    .. math::
        \sum_{i=1}^{N/2} [100(x_{2i-1}^{2} - x_{2i})^2 + (x_{2i-1} - 1)^2].
    """
    return np.sum(
        100.0 * (x[..., ::2] ** 2.0 - x[..., 1::2]) ** 2.0
        + (x[..., ::2] - 1.0) ** 2.0,
        axis=-1,
    )


def rosenbrock(x: np.ndarray) -> np.ndarray:
    r"""Rosenbrock function in N dimensions.

    This is the more involved variant given by

    .. math::
        \sum_{i=1}^{N-1} [100(x_{i+1} - x_{i}^{2})^2 + (1 - x_{i})^2].
    """
    return np.sum(
        100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0
        + (1.0 - x[..., :-1]) ** 2.0,
        axis=-1,
    )


class Rosenbrock(UniformPriorMixin, NDimensionalModel):
    """An n-dimensional Rosenbrock likelihood.

    Defaults to two dimensions and priors defined on [-5, 5]^n.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    uncouple : bool
        Enable the uncoupled (simpler) version of the Rosenbrock likelihood.
    """

    def __init__(
        self,
        dims: int = 2,
        bounds: Union[Sequence[int], np.ndarray] = [-5.0, 5.0],
        uncoupled: float = False,
    ) -> None:
        super().__init__(dims, bounds)

        self.uncoupled = uncoupled
        if self.uncoupled:
            self._fn = uncoupled_rosenbrock
        else:
            self._fn = rosenbrock

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Rosenbrock Log-likelihood."""
        return -self._fn(self.unstructured_view(x))



"""
N-dimensional Gaussian likelihood
"""
from typing import Sequence, Union

import numpy as np
from scipy.stats import halfnorm

from .base import NDimensionalModel, UniformPriorMixin
from .gaussian import compute_gaussian_ln_evidence


class HalfGaussian(UniformPriorMixin, NDimensionalModel):
    """A multi-dimensional Half Gaussian likelihood.

    The lower bound must be set to zero in all dimensions.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    """

    def __init__(
        self,
        dims: int = 2,
        bounds: Union[Sequence[int], np.ndarray] = [0.0, 10.0],
    ) -> None:
        super().__init__(dims, bounds)
        if not all(self.lower_bounds == 0.0):
            raise ValueError("Lower bounds must all be zero!")
        self.ln_evidence = compute_gaussian_ln_evidence(bounds, dims=self.dims)

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Gaussian log-likelihood."""
        log_l = np.zeros(x.size)
        for n in self.names:
            log_l += halfnorm.logpdf(x[n])
        return log_l



# -*- coding: utf-8 -*-
"""
N-dimensional pyramid likelihood
"""
from typing import Sequence, Union

import numpy as np
from .base import NDimensionalModel, UniformPriorMixin


class Pyramid(UniformPriorMixin, NDimensionalModel):
    """N-dimensional pyramid likelihood with uniform priors.

    Parameters
    ----------
    dims :
        Number of dimensions.
    bounds :
        Prior bounds.
    """

    def __init__(
        self,
        dims: int = 2,
        bounds: Union[Sequence[int], np.ndarray] = [-10, 10],
    ) -> None:
        super().__init__(dims, bounds)

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Compute the log-likelihood.

        Parameters
        ----------
        x :
            Point or array of points as a structured array with fields
            that match the names of the model.

        Returns
        -------
        numpy.ndarray
            One-dimensional array of log-likelihoods
        """
        return -np.sum(np.abs(self.unstructured_view(x)), axis=-1)



"""Signal plus noise models."""
from abc import abstractmethod
from typing import Dict, List, Optional

import numpy as np

from .base import BaseModel, UniformPriorMixin


class GaussianNoisePlusSignal(UniformPriorMixin, BaseModel):
    """Gaussian noise plus signal model.

    Parameters
    ----------
    names : List[str]
        Names of the parameters
    truth : dict
        Dictionary contain the true value for the injected signal.
    sigma : float
        Standard deviation of the Gaussian noise.
    bounds : Dict
        Prior bounds for the parameters.
    n_points : int
        The number of data points to use.
    start : float
        The starting x-value.
    end : float
        The ending x-value.
    """

    def __init__(
        self,
        names: List[str],
        truth: Optional[Dict] = None,
        sigma: float = 1.0,
        bounds: Dict = None,
        n_points: int = 100,
        start: float = 0.0,
        end: float = 1.0,
    ) -> None:
        self.names = names

        if truth is None:
            truth = {k: np.random.uniform(*v) for k, v in bounds.items()}
        elif list(truth.keys()) != self.names:
            raise ValueError("Keys in truth dictionary do not match names")

        self.bounds = bounds
        self.truth = truth
        self.sigma = sigma

        self.x = np.linspace(start, end, n_points)[:, np.newaxis]
        self.data = self.signal_model(
            **self.truth
        ) + self.sigma * np.random.randn(n_points, 1)

    @abstractmethod
    def signal_model(self):
        """Must be implemented by the child class.

        Should be defined using named arguments.
        """
        raise NotImplementedError

    def log_likelihood(self, x: np.ndarray) -> np.ndarray:
        """Compute the log-likelihood"""
        fits = self.signal_model(**{n: x[n] for n in self.names})
        log_l = np.sum(
            -0.5 * (((self.data - fits) / self.sigma) ** 2)
            - np.log(2 * np.pi * self.sigma**2),
            axis=0,
        )
        return log_l


class LinearSignal(GaussianNoisePlusSignal):
    """Linear signal model in Gaussian noise.

    Parameter names are: m, c
    """

    def __init__(
        self,
        truth: Optional[Dict] = None,
        sigma: float = 1,
        bounds: Optional[Dict] = None,
        n_points: int = 100,
        start: float = 0,
        end: float = 10,
    ) -> None:
        names = ["m", "c"]
        if bounds is None:
            bounds = dict(
                m=[-1, 1],
                c=[-1, 1],
            )

        super().__init__(
            names=names,
            truth=truth,
            sigma=sigma,
            bounds=bounds,
            n_points=n_points,
            start=start,
            end=end,
        )

    def signal_model(self, *, m, c) -> np.ndarray:
        """Linear signal model."""
        return m * self.x + c


class SinusoidalSignal(GaussianNoisePlusSignal):
    """Sinusoidal signal model in Gaussian noise.

    Parameter names are: amp, phase, f, offset
    """

    def __init__(
        self,
        truth: Optional[Dict] = None,
        sigma: float = 1,
        bounds: Optional[Dict] = None,
        n_points: int = 100,
        start: float = 0,
        end: float = 10,
    ) -> None:
        names = ["amp", "phase", "f", "offset"]
        if bounds is None:
            bounds = dict(
                amp=[0, 1],
                f=[0, 5],
                phase=[0, 2 * np.pi],
                offset=[0, 5],
            )

        super().__init__(
            names=names,
            truth=truth,
            sigma=sigma,
            bounds=bounds,
            n_points=n_points,
            start=start,
            end=end,
        )

    def signal_model(self, *, amp, f, phase, offset) -> np.ndarray:
        """Sinusoidal signal model."""
        return amp * np.sin(2 * np.pi * f * self.x + phase) + offset



# -*- coding: utf-8 -*-
"""
Specific tests for the Pyramid model.
"""
from nessai.livepoint import parameters_to_live_point
import pytest

from nessai_models.pyramid import Pyramid


@pytest.mark.parametrize("dims", [2, 4, 8, 16])
@pytest.mark.integration_test
def test_max_log_likelihood(dims):
    """Assert the maximum log-likelihood is zero at [0, 0]"""
    model = Pyramid(dims=dims)
    x = parameters_to_live_point(dims * [0.0], [f"x{i}" for i in range(dims)])
    assert model.log_likelihood(x) == 0.0



# -*- coding: utf-8 -*-
"""
Base models that remove the need to repeat code between models.
"""
from typing import Sequence, Union

from nessai.model import Model
import numpy as np


class BaseModel(Model):
    """Model that includes an evidence attribute.

    Attributes
    ----------
    ln_evidence : float
        Natural log-evidence. Not set by default.
    """

    ln_evidence: float = None


class NDimensionalModel(BaseModel):
    """Model with basic init for n-dimensional likelihoods.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    bounds : Union[Sequence[float], numpy.ndarray]
        Prior bounds.
    """

    def __init__(
        self, dims: int, bounds: Union[Sequence[float], np.ndarray]
    ) -> None:
        self.names = [f"x_{i}" for i in range(dims)]
        if isinstance(bounds, (Sequence, np.ndarray)):
            if len(bounds) == 2:
                bounds = np.asarray(bounds, dtype=float)
                self.bounds = {n: bounds for n in self.names}
            else:
                raise ValueError("bounds must have length 2.")
        else:
            raise TypeError("Invalid type for `bounds` argument.")


class UniformPriorMixin:
    """Mixin class that defines a uniform prior."""

    def log_prior(self, x: np.ndarray) -> np.ndarray:
        """Log probability for a uniform prior.

        Also checks if samples are within the prior bounds.

        Parameters
        ----------
        x : numpy.ndarray
            Array of samples.

        Returns
        -------
        numpy.ndarray
            Array of log-probabilities.
        """
        with np.errstate(divide="ignore"):
            log_p = np.log(self.in_bounds(x), dtype=float)
        log_p -= np.sum(np.log(self.upper_bounds - self.lower_bounds))
        return log_p

    def to_unit_hypercube(self, x: np.ndarray) -> np.ndarray:
        """Convert the samples to the unit-hypercube.

        Parameters
        ----------
        x : numpy.ndarray
            Array of samples.

        Returns
        -------
        numpy.ndarray
            Array of rescaled samples.
        """
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (x[n] - self.bounds[n][0]) / (
                self.bounds[n][1] - self.bounds[n][0]
            )
        return x_out

    def from_unit_hypercube(self, x: np.ndarray) -> np.ndarray:
        """Convert samples from the unit-hypercube to the prior space.

        Parameters
        ----------
        x : numpy.ndarray
            Array of samples in the unit-hypercube.

        Returns
        -------
        numpy.ndarray
            Array of sample in the prior space.
        """
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (self.bounds[n][1] - self.bounds[n][0]) * x[
                n
            ] + self.bounds[n][0]
        return x_out



# -*- coding: utf-8 -*-
"""Basic tests for all models."""
import pytest


@pytest.mark.parametrize("n", [1, 10])
def test_model_likelihood_and_prior(ModelClass, n):
    """Check the log-prior and log-likelihood integration.

    Asserts values are returned and the correct size.
    """
    model = ModelClass()
    x = model.new_point(n)
    log_p = model.log_prior(x)
    log_l = model.log_likelihood(x)
    assert log_p.size == n
    assert log_l.size == n
